@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2016 Junjiro R. Okajima
+ * Copyright (C) 2005-2017 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -99,17 +99,15 @@ int aufs_release_nondir(struct inode *inode __maybe_unused, struct file *file)
 {
 	struct au_finfo *finfo;
 	aufs_bindex_t bindex;
-	int delayed;
 
 	finfo = au_fi(file);
-	au_sphl_del(&finfo->fi_hlist,
-		    &au_sbi(file->f_path.dentry->d_sb)->si_files);
+	au_hbl_del(&finfo->fi_hlist,
+		   &au_sbi(file->f_path.dentry->d_sb)->si_files);
 	bindex = finfo->fi_btop;
 	if (bindex >= 0)
 		au_set_h_fptr(file, bindex, NULL);
 
-	delayed = (current->flags & PF_KTHREAD) || in_interrupt();
-	au_finfo_fin(file, delayed);
+	au_finfo_fin(file);
 	return 0;
 }
 
@@ -142,12 +140,12 @@ static int aufs_flush_nondir(struct file *file, fl_owner_t id)
  */
 
 /* Callers should call au_read_post() or fput() in the end */
-struct file *au_read_pre(struct file *file, int keep_fi)
+struct file *au_read_pre(struct file *file, int keep_fi, unsigned int lsc)
 {
 	struct file *h_file;
 	int err;
 
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0);
+	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0, lsc);
 	if (!err) {
 		di_read_unlock(file->f_path.dentry, AuLock_IR);
 		h_file = au_hf_top(file);
@@ -168,6 +166,10 @@ static void au_read_post(struct inode *inode, struct file *h_file)
 }
 
 struct au_write_pre {
+	/* input */
+	unsigned int lsc;
+
+	/* output */
 	blkcnt_t blks;
 	aufs_bindex_t btop;
 };
@@ -183,9 +185,13 @@ static struct file *au_write_pre(struct file *file, int do_ready,
 	struct file *h_file;
 	struct dentry *dentry;
 	int err;
+	unsigned int lsc;
 	struct au_pin pin;
 
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
+	lsc = 0;
+	if (wpre)
+		lsc = wpre->lsc;
+	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1, lsc);
 	h_file = ERR_PTR(err);
 	if (unlikely(err))
 		goto out;
@@ -227,12 +233,11 @@ static void au_write_post(struct inode *inode, struct file *h_file,
 	h_inode = file_inode(h_file);
 	inode->i_mode = h_inode->i_mode;
 	ii_write_unlock(inode);
-	fput(h_file);
-
 	/* AuDbg("blks %llu, %llu\n", (u64)blks, (u64)h_inode->i_blocks); */
 	if (written > 0)
 		au_fhsm_wrote(inode->i_sb, wpre->btop,
 			      /*force*/h_inode->i_blocks > wpre->blks);
+	fput(h_file);
 }
 
 static ssize_t aufs_read(struct file *file, char __user *buf, size_t count,
@@ -247,7 +252,7 @@ static ssize_t aufs_read(struct file *file, char __user *buf, size_t count,
 	sb = inode->i_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
 
-	h_file = au_read_pre(file, /*keep_fi*/0);
+	h_file = au_read_pre(file, /*keep_fi*/0, /*lsc*/0);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
 		goto out;
@@ -297,6 +302,7 @@ static ssize_t aufs_write(struct file *file, const char __user *ubuf,
 	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
+	wpre.lsc = 0;
 	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
@@ -356,7 +362,7 @@ static ssize_t aufs_read_iter(struct kiocb *kio, struct iov_iter *iov_iter)
 	sb = inode->i_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
 
-	h_file = au_read_pre(file, /*keep_fi*/1);
+	h_file = au_read_pre(file, /*keep_fi*/1, /*lsc*/0);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
 		goto out;
@@ -391,6 +397,7 @@ static ssize_t aufs_write_iter(struct kiocb *kio, struct iov_iter *iov_iter)
 	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
+	wpre.lsc = 0;
 	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
@@ -418,7 +425,7 @@ static ssize_t aufs_splice_read(struct file *file, loff_t *ppos,
 	sb = inode->i_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
 
-	h_file = au_read_pre(file, /*keep_fi*/0);
+	h_file = au_read_pre(file, /*keep_fi*/0, /*lsc*/0);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
 		goto out;
@@ -445,6 +452,7 @@ aufs_splice_write(struct pipe_inode_info *pipe, struct file *file, loff_t *ppos,
 	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
+	wpre.lsc = 0;
 	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
@@ -470,6 +478,7 @@ static long aufs_fallocate(struct file *file, int mode, loff_t offset,
 	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
+	wpre.lsc = 0;
 	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
@@ -486,6 +495,88 @@ out:
 	return err;
 }
 
+static ssize_t aufs_copy_file_range(struct file *src, loff_t src_pos,
+				    struct file *dst, loff_t dst_pos,
+				    size_t len, unsigned int flags)
+{
+	ssize_t err;
+	struct au_write_pre wpre;
+	enum { SRC, DST };
+	struct {
+		struct inode *inode;
+		struct file *h_file;
+		struct super_block *h_sb;
+	} a[2];
+#define a_src	a[SRC]
+#define a_dst	a[DST]
+
+	err = -EINVAL;
+	a_src.inode = file_inode(src);
+	if (unlikely(!S_ISREG(a_src.inode->i_mode)))
+		goto out;
+	a_dst.inode = file_inode(dst);
+	if (unlikely(!S_ISREG(a_dst.inode->i_mode)))
+		goto out;
+
+	au_mtx_and_read_lock(a_dst.inode);
+	/*
+	 * in order to match the order in di_write_lock2_{child,parent}(),
+	 * use f_path.dentry for this comparision.
+	 */
+	if (src->f_path.dentry < dst->f_path.dentry) {
+		a_src.h_file = au_read_pre(src, /*keep_fi*/1, AuLsc_FI_1);
+		err = PTR_ERR(a_src.h_file);
+		if (IS_ERR(a_src.h_file))
+			goto out_si;
+
+		wpre.lsc = AuLsc_FI_2;
+		a_dst.h_file = au_write_pre(dst, /*do_ready*/1, &wpre);
+		err = PTR_ERR(a_dst.h_file);
+		if (IS_ERR(a_dst.h_file)) {
+			au_read_post(a_src.inode, a_src.h_file);
+			goto out_si;
+		}
+	} else {
+		wpre.lsc = AuLsc_FI_1;
+		a_dst.h_file = au_write_pre(dst, /*do_ready*/1, &wpre);
+		err = PTR_ERR(a_dst.h_file);
+		if (IS_ERR(a_dst.h_file))
+			goto out_si;
+
+		a_src.h_file = au_read_pre(src, /*keep_fi*/1, AuLsc_FI_2);
+		err = PTR_ERR(a_src.h_file);
+		if (IS_ERR(a_src.h_file)) {
+			au_write_post(a_dst.inode, a_dst.h_file, &wpre,
+				      /*written*/0);
+			goto out_si;
+		}
+	}
+
+	err = -EXDEV;
+	a_src.h_sb = file_inode(a_src.h_file)->i_sb;
+	a_dst.h_sb = file_inode(a_dst.h_file)->i_sb;
+	if (unlikely(a_src.h_sb != a_dst.h_sb)) {
+		AuDbgFile(src);
+		AuDbgFile(dst);
+		goto out_file;
+	}
+
+	err = vfsub_copy_file_range(a_src.h_file, src_pos, a_dst.h_file,
+				    dst_pos, len, flags);
+
+out_file:
+	au_write_post(a_dst.inode, a_dst.h_file, &wpre, err);
+	fi_read_unlock(src);
+	au_read_post(a_src.inode, a_src.h_file);
+out_si:
+	si_read_unlock(a_dst.inode->i_sb);
+	inode_unlock(a_dst.inode);
+out:
+	return err;
+#undef a_src
+#undef a_dst
+}
+
 /* ---------------------------------------------------------------------- */
 
 /*
@@ -576,7 +667,7 @@ static int aufs_mmap(struct file *file, struct vm_area_struct *vma)
 	 *			 au_flag_conv(vma->vm_flags));
 	 */
 	if (!err)
-		err = h_file->f_op->mmap(h_file, vma);
+		err = call_mmap(h_file, vma);
 	if (!err) {
 		au_vm_prfile_set(vma, file);
 		fsstack_copy_attr_atime(inode, file_inode(h_file));
@@ -615,6 +706,7 @@ static int aufs_fsync_nondir(struct file *file, loff_t start, loff_t end,
 	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
+	wpre.lsc = 0;
 	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
@@ -639,7 +731,7 @@ static int aufs_fasync(int fd, struct file *file, int flag)
 	sb = file->f_path.dentry->d_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
 
-	h_file = au_read_pre(file, /*keep_fi*/0);
+	h_file = au_read_pre(file, /*keep_fi*/0, /*lsc*/0);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
 		goto out;
@@ -662,12 +754,13 @@ static int aufs_setfl(struct file *file, unsigned long arg)
 	sb = file->f_path.dentry->d_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
 
-	h_file = au_read_pre(file, /*keep_fi*/0);
+	h_file = au_read_pre(file, /*keep_fi*/0, /*lsc*/0);
 	err = PTR_ERR(h_file);
 	if (IS_ERR(h_file))
 		goto out;
 
-	arg |= vfsub_file_flags(file) & FASYNC; /* stop calling h_file->fasync */
+	/* stop calling h_file->fasync */
+	arg |= vfsub_file_flags(file) & FASYNC;
 	err = setfl(/*unused fd*/-1, h_file, arg);
 	fput(h_file); /* instead of au_read_post() */
 
@@ -719,5 +812,6 @@ const struct file_operations aufs_file_fop = {
 	.aio_splice_write = aufs_aio_splice_write,
 	.aio_splice_read  = aufs_aio_splice_read,
 #endif
-	.fallocate	= aufs_fallocate
+	.fallocate	= aufs_fallocate,
+	.copy_file_range = aufs_copy_file_range
 };
