@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2016 Junjiro R. Okajima
+ * Copyright (C) 2005-2017 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,29 +19,36 @@
  * sub-routines for VFS
  */
 
+#include <linux/mnt_namespace.h>
 #include <linux/namei.h>
 #include <linux/nsproxy.h>
 #include <linux/security.h>
 #include <linux/splice.h>
-#ifdef CONFIG_AUFS_BR_FUSE
-#include "../fs/mount.h"
-#endif
 #include "aufs.h"
 
 #ifdef CONFIG_AUFS_BR_FUSE
 int vfsub_test_mntns(struct vfsmount *mnt, struct super_block *h_sb)
 {
-	struct nsproxy *ns;
-
 	if (!au_test_fuse(h_sb) || !au_userns)
 		return 0;
 
-	ns = current->nsproxy;
-	/* no {get,put}_nsproxy(ns) */
-	return real_mount(mnt)->mnt_ns == ns->mnt_ns ? 0 : -EACCES;
+	return is_current_mnt_ns(mnt) ? 0 : -EACCES;
 }
 #endif
 
+int vfsub_sync_filesystem(struct super_block *h_sb, int wait)
+{
+	int err;
+
+	lockdep_off();
+	down_read(&h_sb->s_umount);
+	err = __sync_filesystem(h_sb, wait);
+	up_read(&h_sb->s_umount);
+	lockdep_on();
+
+	return err;
+}
+
 /* ---------------------------------------------------------------------- */
 
 int vfsub_update_h_iattr(struct path *h_path, int *did)
@@ -59,7 +66,7 @@ int vfsub_update_h_iattr(struct path *h_path, int *did)
 	h_sb = h_path->dentry->d_sb;
 	*did = (!au_test_fs_remote(h_sb) && au_test_fs_refresh_iattr(h_sb));
 	if (*did)
-		err = vfs_getattr(h_path, &st);
+		err = vfsub_getattr(h_path, &st);
 
 	return err;
 }
@@ -585,6 +592,7 @@ int vfsub_iterate_dir(struct file *file, struct dir_context *ctx)
 	lockdep_on();
 	if (err >= 0)
 		vfsub_update_h_iattr(&file->f_path, /*did*/NULL); /*ignore*/
+
 	return err;
 }
 
